
/*     *******************************************************************************
	XVisual GUI Development System for Linux/X-Window
	Copyright (C) 1996 Steve Carrie.

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
        ******************************************************************************* */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xaw/Text.h>
#include <X11/xv/xvCore.h>
#include <X11/xv/xvBasicBitmap.h>
#include <X11/xv/xvFont.h>
#include <X11/xv/xvContext.h>
#include <X11/xv/xvForm.h>
#include <X11/xv/xvLabel.h>
#include <X11/xv/xvButton.h>
#include <X11/xv/xvView.h>
#include <X11/xv/xvList.h>
#include <X11/xv/xvToggle.h>
#include <X11/xv/xvEditField.h>
#include <X11/xv/xvPopupShell.h>
#include <X11/xv/xvMsgBox.h>
#include <X11/xv/xvFileDialog.h>

// ------------------------------------------

xvFileDialog::xvFileDialog()
{
	numfiles=0;
	numdirs=0;
	currentpath[0]='\0';
	currentfile[0]='\0';
	allocflag=0;	
	hiddenflag=0;
}

// ------------------------------------------

xvFileDialog::xvFileDialog(String name, xvCore & parent, String pathname,
				XtCallbackProc filelist_callback, XtCallbackProc dirlist_callback,
				XtCallbackProc ok_callback, XtCallbackProc cancel_callback,
				XtCallbackProc err_callback, XtActionProc edit_callback,
				xvContext & appcontext, String mask, Boolean maskActive,
				Boolean allowDirchange)
{
	init(name, parent, pathname, filelist_callback, dirlist_callback, ok_callback,
		cancel_callback, err_callback, edit_callback, appcontext, mask, maskActive,
		allowDirchange);
}

// ------------------------------------------

xvFileDialog::~xvFileDialog()
{
	if (allocflag != 0) {
		delete filelist;
		delete dirlist;
		delete superlist;
		delete statlist;
		allocflag=0;
	}
}

// ------------------------------------------

void xvFileDialog::init(String name, xvCore & parent, String pathname,
				XtCallbackProc filelist_callback, XtCallbackProc dirlist_callback,
				XtCallbackProc ok_callback, XtCallbackProc cancel_callback,
				XtCallbackProc err_callback, XtActionProc edit_callback,
				xvContext & appcontext, String mask,  Boolean maskActive,
				Boolean allowDirchange)
{
	numfiles=0;
	numdirs=0;
	currentpath[0]='\0';
	currentfile[0]='\0';
	allocflag=0;	
	hiddenflag=0;

	usemask=maskActive;
	tform.init(name, parent, TRUE, TRUE);
	w=tform.ID();

//      ** Fileeditbox label
	tflabel.init("Filename", tform);
	tflabel.SetBorderWidth(0);
	tflabel.SetHorizDistance(25);
	tflabel.SetVertDistance(25);

//     ** file edit box
	tfedit.init("edit", tform, XawtextEdit, 20, 150);
	tfedit.SetRelVertical(tflabel);
	tfedit.SetHorizDistance(25);	
	tfedit.SetVertDistance(3);
	XtTranslations trans=XtParseTranslationTable(
		"<Key>Return: select-all() fddataentry() \n\
		!Meta<Key> i: no-op()");
	tfedit.OverrideTranslations(trans);
	appcontext.AddAction("fddataentry", edit_callback);
	UpdateFileBox(mask);
	strcpy(filemask, mask);

//      ** file list label
	tfblabel.init("Files           ", tform);
	tfblabel.SetBorderWidth(0);
	tfblabel.SetRelVertical(tfedit);
	tfblabel.SetHorizDistance(25);
	tfblabel.SetVertDistance(25);
	tfblabel.SetWidth(150);

//     ** file list view
	tfbview.init("fileview", tform, TRUE, FALSE, TRUE);
	tfbview.SetRelVertical(tfblabel);
	tfbview.SetHorizDistance(25);
	tfbview.SetVertDistance(3);

//	** file list	
	tfblist.init("filelist", tfbview, filelist_callback,  numfiles, filelist, 1);
	tfblist.SetWidth(150);
	
	tfbview.SetHeight(183);
	tfbview.SetWidth(150);	

//     ** dir list label
	tdlabel.init("Directories", tform);
	tdlabel.SetBorderWidth(0);
	tdlabel.SetRelHorizontal(tfbview);
	tdlabel.SetHorizDistance(25);
	tdlabel.SetVertDistance(25);

//     ** dir path
	pathlabel.init("  ", tform);
	pathlabel.SetBorderWidth(0);
	pathlabel.SetRelVertical(tdlabel);
	pathlabel.SetRelHorizontal(tfbview);
	pathlabel.SetHorizDistance(25);
	pathlabel.SetVertDistance(15);
	pathlabel.SetJustify(XtJustifyLeft);

//     ** dir view
	tdview.init("fileview", tform, TRUE, FALSE, TRUE);
	tdview.SetRelVertical(pathlabel);
	tdview.SetRelHorizontal(tfbview);
	tdview.SetHorizDistance(25);
	tdview.SetVertDistance(15);
	if (! allowDirchange)
		tdview.SetSensitive(FALSE);

//      ** dirlist	
	tdlist.init("dirlist", tdview, dirlist_callback, numdirs, dirlist, 1);
	tdlist.SetWidth(150);

	tdview.SetHeight(205);
	tdview.SetWidth(150);

//     ** OK button
	ok.init("OK", tform, ok_callback);
	ok.SetRelHorizontal(tdview);
	ok.SetVertDistance(50);
	ok.SetHorizDistance(25);
	ok.SetWidth(100);
	ok.SetBeNiceToColormap(FALSE);

//     ** Cancel Button
	cancel.init("Cancel", tform, cancel_callback);
	cancel.SetRelVertical(ok);
	cancel.SetRelHorizontal(tdview);
	cancel.SetVertDistance(25);
	cancel.SetHorizDistance(25);
	cancel.SetWidth(100);
	cancel.SetBeNiceToColormap(FALSE);

//     ** Toggle
	ttoggle.init("Show Hidden Dirs/Files", tform, (XtCallbackProc) ToggleClick, NULL);
	ttoggle.SetRelVertical(tfbview);
	ttoggle.SetVertDistance(25);
	ttoggle.SetHorizDistance(25);
	ttoggle.SetBeNiceToColormap(FALSE);

//     ** rightside & bottomside spacers
	rspacer.init(" ", tform);
	rspacer.SetRelHorizontal(ok);
	rspacer.SetHorizDistance(5);
	rspacer.SetBorderWidth(0);
	bspacer.init(" ", tform);
	bspacer.SetRelVertical(ttoggle);
	bspacer.SetVertDistance(5);
	bspacer.SetBorderWidth(0);

	tmsg.init("Protected", pathlabel,
			"DIRECTORY ACCESS ERROR\n\nAccess Denied.",
			XVMSGSIMPLEWARN, "SkyBlue",
			err_callback);

	NewPath(pathname);
}

// ---------------------------------------------

void xvFileDialog::ToggleClick(Widget wd, XtPointer cl, XtPointer ca)
{

}

// ---------------------------------------------

String xvFileDialog::File(XtPointer call)
{
		
	XawListReturnStruct *xr=(XawListReturnStruct *) call;
	
	UpdateFileBox(filelist[xr->list_index]);
	return(currentfile);
}

// --------------------------------------------

String xvFileDialog::Directory(XtPointer call)
{
	char buf[180];

	XawListReturnStruct *xr=(XawListReturnStruct *) call;
	
	if (strcmp(xr->string, "..")==0) {
		strcpy(buf, currentpath);
		int i=strlen(buf);
		if (i>=2) {
			i--;
			while ((buf[i]!='/') && (i>0))
				i--;		
			buf[i]='\0';
		}
		if (i<2)
			if (buf[0]!='/') 
				strcpy(buf, "/");
	}
	else {
		strcpy(buf, currentpath);
		if (buf[strlen(buf)-1]!='/')
			strcat(buf, "/");
		strcat(buf, xr->string);
	}
	NewPath(buf);
}

// --------------------------------------------

void xvFileDialog::UpdateFileBox(String newstr)
{
	struct XawTextBlock textdata;

	int l=strlen(currentfile);
	textdata.firstPos=0;
	textdata.length=strlen(newstr);
	textdata.ptr=newstr;
	textdata.format=FMT8BIT;
	tfedit.Replace(0, l, &textdata);
	strcpy(currentfile, newstr);
}

// ---------------------------------------------

String xvFileDialog::SelectedFile(String returnpathname)
{
	char *s;
	XtVaGetValues(tfedit.ID(), XtNstring, &s, NULL);
	strcpy(currentfile, s);
	strcpy(returnpathname, currentpath);
	strcat(returnpathname, "/");
	strcat(returnpathname, currentfile);

	return(currentfile);
}

// ----------------------------------------------

void xvFileDialog::NewPath(String pathname)
{
	int numentries=GetNumEntries(pathname);
	if (numentries >=0 ) {
		if (allocflag != 0) {
			delete filelist;
			delete dirlist;
			delete superlist;
			delete statlist;
			allocflag=0;
		}

		ParseFileTree(pathname, numentries);
		UpdateFileList();
		UpdateDirList();
//		UpdateFileBox(" ");
		strcpy(currentpath, pathname);
	
		sprintf(currentnum, "Files: %d    ", numfiles);
		tfblabel.SetLabel(currentnum);	
		pathlabel.SetLabel(currentpath);
		pathlabel.SetWidth(150);
	}
	else {
		Position npx=0, npy=0;
		Position vx=tform.GetWidth()/4;
		Position vy=tform.GetHeight()/4;
		XtTranslateCoords(tform.ID(), vx, vy, &npx, &npy); 
		tmsg.Open(npx,npy);
	}
}

// ----------------------------------------------

void xvFileDialog::ClearError(XtPointer calldata)
{
	tmsg.Close();
}

// ----------------------------------------------

void xvFileDialog::UpdateFileList()
{
	tfblist.SetList(numfiles, filelist);
}

// ----------------------------------------------

void xvFileDialog::UpdateDirList()
{
	tdlist.SetList(numdirs, dirlist);
}

// --------------------------------------

int xvFileDialog::GetNumEntries(String pathname)
{
	DIR *dp;
	struct dirent *d;
	int cnt=0;

	dp=opendir(pathname);
	if (dp!=NULL) { 
		while ((d=readdir(dp)) !=NULL) 
			cnt++;
		closedir(dp);
	}
	else
		cnt=-1;

	return(cnt);
}

// ------------------------------------

int xvFileDialog::ParseFileTree(String pathname, int entries)
{
	DIR *dp;
	int listed=0;
	char buf[180];
	struct dirent *d;
	hiddenflag=(int) ttoggle.GetState();
	
	dp=opendir(pathname);
	if (dp!=NULL) {
		allocflag=1;
		superlist= new struct dirent[entries];
		statlist=new struct stat[entries];
		dirlist=new char*[entries];
		filelist=new char*[entries];
		for (int i=0; i<entries; i++) {
			dirlist[i]=NULL;
			filelist[i]=NULL;
		}
			
		numfiles=0;
		numdirs=0;
		
		while ((d=readdir(dp)) != NULL) {
			strcpy(superlist[listed].d_name,d->d_name);
			strcpy(buf, pathname);
			strcat(buf, "/");
			strcat(buf, superlist[listed].d_name);
			stat(buf, &statlist[listed]);
	
			if (S_ISDIR(statlist[listed].st_mode)!=0) {
				if (strcmp(superlist[listed].d_name, ".")!=0) 
					if ((hiddenflag != 0) || (superlist[listed].d_name[0] != '.') || (superlist[listed].d_name[1] =='.'))
						dirlist[numdirs++]=superlist[listed].d_name;
			}
			else	
				if ((hiddenflag != 0) || (superlist[listed].d_name[0] != '.'))
					if ((! usemask) || ((usemask) && (strstr(superlist[listed].d_name, filemask) != NULL)))
						filelist[numfiles++]=superlist[listed].d_name;
				
			listed++;
			
		}					
		closedir(dp);
		
	}
	
	return(listed);
}

// ---------------------------------------

void xvFileDialog::EditEntry()
{
	XawTextPosition start, end;
	XawTextBlock block;

	tfedit.Position(&start, &end);
	tfedit.SourceRead(0, &block, 40);
	strncpy(currentfile, block.ptr, end);
	currentfile[end]='\0';
	tfedit.UnsetSelection();
}
	
// ------------------------------------------------------------
